#!/usr/local/bin python3
# -*- coding: utf-8 -*-

"""
    created by FAST-DEV 2021/4/6
"""

import inspect

from fast_tracker import Layer, ComponentType, config
from fast_tracker.trace import tags
from fast_tracker.trace.carrier import Carrier
from fast_tracker.trace.context import get_context
from fast_tracker.trace.tags import Tag
from fast_tracker.utils import functions


def install():
    from http.server import BaseHTTPRequestHandler

    _handle = BaseHTTPRequestHandler.handle

    def _fast_handle(handler: BaseHTTPRequestHandler):
        clazz = handler.__class__
        if 'werkzeug.serving.WSGIRequestHandler' == ".".join([clazz.__module__, clazz.__name__]):
            wrap_werkzeug_request_handler(handler)
        else:
            wrap_default_request_handler(handler)
        _handle(handler)

    BaseHTTPRequestHandler.handle = _fast_handle

    def _fast_send_response_only(self, code, *args, **kwargs):
        self._status_code = code

        return _send_response_only(self, code, *args, **kwargs)

    _send_response_only = BaseHTTPRequestHandler.send_response_only
    BaseHTTPRequestHandler.send_response_only = _fast_send_response_only


def wrap_werkzeug_request_handler(handler):
    """
    Wrap run_wsgi of werkzeug.serving.WSGIRequestHandler to add fast_tracker instrument code.
    """
    _run_wsgi = handler.run_wsgi

    def _wrap_run_wsgi():
        context = get_context()
        carrier = Carrier()
        trace_id_name = config.get_trace_id_name()
        for item in carrier:
            if trace_id_name.capitalize() in handler.headers and type(item) is Carrier:
                item.set_frontend_trace_id(handler.headers[trace_id_name.capitalize()])
            item.val = handler.headers[item.key.capitalize()]
        path = handler.path or '/'
        functions.log("wsgi path:%s,host:%s", handler.path, handler.headers["Host"])
        with context.new_entry_span(op=path.split("?")[0], carrier=carrier) as span:
            url = 'http://' + handler.headers["Host"] + path if 'Host' in handler.headers else path
            span.layer = Layer.Http
            span.component = ComponentType.General
            span.peer = '%s:%s' % handler.client_address
            span.tag(Tag(key=tags.HttpMethod, val=handler.command))
            span.tag(Tag(key=tags.HttpUrl, val=url))
            # HttpPath 只需要获取PATH_INFO信息即可，不需要？后的内容
            span.tag(Tag(key=tags.HttpPath, val=path.split("?")[0]))
            status_code = int(getattr(handler, '_status_code', -1))
            span.tag(Tag(key=tags.HttpStatus, val=status_code))

            try:
                return _run_wsgi()
            finally:
                status_code = int(getattr(handler, '_status_code', -1))
                if status_code > -1:
                    span.tag(Tag(key=tags.HttpStatus, val=status_code, overridable=True))
                    if status_code >= 400:
                        span.error_occurred = True

    handler.run_wsgi = _wrap_run_wsgi

    def _fast_send_response(self, code, *args, **kwargs):
        self._status_code = code

        return _send_response(self, code, *args, **kwargs)

    WSGIRequestHandler = handler.__class__

    if not getattr(WSGIRequestHandler, '_fast_wrapped', False):
        _send_response = WSGIRequestHandler.send_response
        WSGIRequestHandler.send_response = _fast_send_response
        WSGIRequestHandler._fast_wrapped = True


def wrap_default_request_handler(handler):
    http_methods = ('GET', 'HEAD', 'POST', 'PUT', 'DELETE', 'CONNECT', 'OPTIONS', 'TRACE', 'PATCH')
    for method in http_methods:
        _wrap_do_method(handler, method)


def _wrap_do_method(handler, method):
    if hasattr(handler, 'do_' + method) and inspect.ismethod(getattr(handler, 'do_' + method)):
        _do_method = getattr(handler, 'do_' + method)

        def _fast_do_method():
            context = get_context()
            carrier = Carrier()
            trace_id_name = config.get_trace_id_name()
            for item in carrier:
                item.val = handler.headers[item.key.capitalize()]
                if trace_id_name.capitalize() in handler.headers and type(item) is Carrier:
                    item.set_frontend_trace_id(handler.headers[trace_id_name.capitalize()])
            path = handler.path or '/'
            functions.log("wrap path:%s", path)
            with context.new_entry_span(op=path.split("?")[0], carrier=carrier) as span:
                url = 'http://' + handler.headers["Host"] + path if 'Host' in handler.headers else path
                span.layer = Layer.Http
                span.component = ComponentType.General
                span.peer = '%s:%s' % handler.client_address
                span.tag(Tag(key=tags.HttpMethod, val=method))
                span.tag(Tag(key=tags.HttpUrl, val=url))

                try:
                    _do_method()
                finally:
                    status_code = int(getattr(handler, '_status_code', -1))
                    if status_code > -1:
                        span.tag(Tag(key=tags.HttpStatus, val=status_code, overridable=True))
                        if status_code >= 400:
                            span.error_occurred = True

        setattr(handler, 'do_' + method, _fast_do_method)



#!/usr/local/bin python3
# -*- coding: utf-8 -*-

"""
    created by FAST-DEV 2021/4/6
"""

from fast_tracker import Layer, ComponentType
from fast_tracker.trace import tags
from fast_tracker.trace.context import get_context
from fast_tracker.trace.tags import Tag
from fast_tracker import config
from fast_tracker.utils import functions


def install():
    from requests import Session

    _request = Session.request

    def _fast_request(
            this: Session,
            method,
            url,
            params=None,
            data=None,
            headers=None,
            cookies=None,
            files=None,
            auth=None,
            timeout=None,
            allow_redirects=True,
            proxies=None,
            hooks=None,
            stream=None,
            verify=None,
            cert=None,
            json=None,
    ):

        from urllib.parse import urlparse

        url_param = urlparse(url)

        # ignore trace fast_tracker self request
        if config.protocol == "http" and config.collector_address.rstrip("/").endswith(url_param.netloc):
            return _request(
                this,
                method,
                url,
                params,
                data,
                headers,
                cookies,
                files,
                auth,
                timeout,
                allow_redirects,
                proxies,
                hooks,
                stream,
                verify,
                cert,
                json,
            )

        context = get_context()
        # functions.log("request url:%s", url_param.path)
        with context.new_exit_span(op=url_param.path.split("?")[0] or "/", peer=url_param.netloc) as span:
            carrier = span.inject()
            functions.log("frq carrier val:%s",carrier.val)
            span.layer = Layer.Http
            span.component = ComponentType.Requests

            if headers is None:
                headers = {}
            for item in carrier:
                headers[item.key] = item.val

            span.tag(Tag(key=tags.HttpMethod, val=method.upper()))
            span.tag(Tag(key=tags.HttpUrl, val=url))
            # HttpPath 只需要获取PATH_INFO信息即可，不需要？后的内容
            span.tag(Tag(key=tags.HttpPath, val=url_param.path))

            res = _request(
                this,
                method,
                url,
                params,
                data,
                headers,
                cookies,
                files,
                auth,
                timeout,
                allow_redirects,
                proxies,
                hooks,
                stream,
                verify,
                cert,
                json,
            )

            span.tag(Tag(key=tags.HttpStatus, val=res.status_code, overridable=True))
            if res.status_code >= 400:
                span.error_occurred = True

            return res

    Session.request = _fast_request



#!/usr/local/bin python3
# -*- coding: utf-8 -*-

"""
    created by FAST-DEV 2021/4/6
"""
import atexit
from queue import Queue, Full
from threading import Thread, Event
from typing import TYPE_CHECKING

from fast_tracker import config, plugins, loggings
from fast_tracker.loggings import logger
from fast_tracker.agent.protocol import Protocol
from fast_tracker.utils import functions

if TYPE_CHECKING:
    from fast_tracker.trace.context import Segment


def __heartbeat():
    while not __finished.is_set():  # 返回Event的内部标识是否为True，为True表示当前线程不阻塞
        if connected():
            __protocol.heartbeat()

        __finished.wait(30 if connected() else 3)


def __report():
    while not __finished.is_set():
        if connected():
            __protocol.report(__queue)  # is blocking actually
        __finished.wait(1)


# __heartbeat_thread = Thread(name='HeartbeatThread', target=__heartbeat, daemon=True)
__report_thread = Thread(name="ReportThread", target=__report, daemon=True)
__queue = Queue(maxsize=10000)
__finished = Event()
__protocol = Protocol()  # type: Protocol
__started = False


def __init():
    """
    初始化插件
    :return:
    """
    global __protocol
    if config.protocol == "udp":
        from fast_tracker.agent.protocol.udp import UdpProtocol

        __protocol = UdpProtocol()

    # plugins文件夹下面的所有插件都初始化（或者说是安装）
    plugins.install()


def __fini():
    __protocol.report(__queue, False)
    __queue.join()


def start():
    functions.log("Config.Enable is %s ", config.enable)
    # 如果探针开关关闭,则不执行探针的初始化操作，不启动探针
    if not config.enable:
        return False

    # if not config.product_code or not config.app_code:
    #     functions.log("缺少product_code或app_code字段值，探针初始化失败")
    #     return False

    flag = False
    try:
        # 猴子补丁，运行时动态替换
        from gevent import monkey

        flag = monkey.is_module_patched("socket")
    except ImportError:  # ImportError python2\3都可以使用，ModuleNotFoundError只能用于3.6+
        functions.log("未发现使用gevent，如果确实未使用gevent，请忽略。")

    if flag:
        import grpc.experimental.gevent as grpc_gevent

        grpc_gevent.init_gevent()
    global __started
    if __started:
        functions.log("此代理只能启动一次")
        return False
    # 初始化log模块
    loggings.init()
    # 配置需要忽略的文件夹
    config.finalize()
    __started = True
    # 插件初始化
    __init()
    # 心跳线程启动
    # __heartbeat_thread.start()
    # 数据上报线程启动
    __report_thread.start()
    # atexit：退出处理器
    atexit.register(__fini)
    # 打印配置
    functions.log(
        "探针enable配置为：%r,  logging_level配置为：%r， logging_filepath配置为：%r",
        config.enable,
        config.tracker_logging.get("Level"),
        config.tracker_logging.get("FilePath"),
    )


def stop():
    atexit.unregister(__fini)
    __fini()
    __finished.set()


def started():
    return __started


def connected():
    return __protocol.connected()


def archive(segment: "Segment"):
    try:  # unlike checking __queue.full() then inserting, this is atomic
        __queue.put(segment, block=False)
    except Full:
        functions.log("队列已满，该segment将被放弃!")
        # logger.warning("队列已满，该segment将被放弃!")



#!/usr/local/bin python3
# -*- coding: utf-8 -*-

"""
    created by FAST-DEV 2021/4/6
"""

from multiprocessing import Process

from fast_tracker import config, agent


class FastProcess(Process):

    def __init__(self, group=None, target=None, name=None, args=(), kwargs={}, *,
                 daemon=None):
        super(FastProcess, self).__init__(group=group, target=target, name=name, args=args, kwargs=kwargs, daemon=daemon)
        self._fast_config = config.serialize()

    def run(self):
        if agent.started() is False:
            config.deserialize(self._fast_config)
            agent.start()
        super(FastProcess, self).run()



#!/usr/local/bin python3
# -*- coding: utf-8 -*-

"""
    created by iprobeyang@gmail.com 2021/4/21
"""

class ReaderType:
    Cookie = "Cookie"
    RequestHeader = "RequestHeader"
    QueryString = "QueryString"
    Environment = "Environment"


#!/usr/local/bin python3
# -*- coding: utf-8 -*-

"""
    created by iprobeyang@gmail.com 2021/4/20
"""

from fast_tracker import Layer, ComponentType, config
from fast_tracker.trace import tags
from fast_tracker.trace.carrier import Carrier
from fast_tracker.trace.context import get_context
from fast_tracker.trace.span import NoopSpan
from fast_tracker.trace.tags import Tag
from fast_tracker.utils.reader_type import ReaderType
from fast_tracker.utils import functions
from fast_tracker.utils import  lang

def install():
    from falcon import API, request, response

    _original_falcon_api = API.__call__
    _original_falcon_handle_exception = API._handle_exception

    def params_tostring(params):
        return "\n".join([k + "=" + v for k, v in params.items()])

    def _fast_falcon_api(this: API, env, start_response):
        context = get_context()
        carrier = Carrier()
        headers = get_headers(env)
        trace_id_name = config.get_trace_id_name()
        trace_name = config.get_tracker_name()
        for item in carrier:
            # functions.log("trace_id_name:%s,trace_name:%s,item.key:%s,type:%s",trace_id_name,trace_name,item.key,type(item))
            if item.key.capitalize() in headers:
                item.val = headers[item.key.capitalize()]
            if item.key.upper() in headers:
                item.val = headers[item.key.upper()]
        if trace_id_name.capitalize() in headers:
            carrier.set_frontend_trace_id(headers[trace_id_name.capitalize()])
        if trace_id_name.upper() in headers:
            carrier.set_frontend_trace_id(headers[trace_id_name.upper()])
        if trace_name.upper() in headers:
            carrier.set_frontend_trace_id(headers[trace_name.upper()])
        if "parent_id" in headers:
            carrier.set_parent_id(headers["parent_id"])
        from falcon import RequestOptions
        req = request.Request(env, RequestOptions())
        with context.new_entry_span(op=functions.getUri(str(req.url.split("?")[0])), carrier=carrier) as span:
            span.layer = Layer.Http
            span.component = ComponentType.Falcon

            from falcon import ResponseOptions

            resp_obj = response.Response(ResponseOptions())

            # resp = _original_falcon_api(this, env, start_response)
            # span.op = str(req.url).split("?")[0]
            span.op = functions.getUri(str(req.url).split("?")[0])
            span.peer = "%s:%s" % (req.remote_addr, req.port)
            span.tag(Tag(key=tags.HttpMethod, val=req.method))
            span.tag(Tag(key=tags.HttpUrl, val=str(req.url)))
            span.tag(Tag(key=tags.HttpPath, val=str(req.url).split("?")[0]))


            if req.params:
                span.tag(
                    Tag(key=tags.HttpParams, val=params_tostring(req.params)[0:])
                )

            set_code(config.tenant_code_reader, req, "tenant_code")
            set_code(config.user_code_reader, req, "user_code")
            set_code(config.service_version_reader,req,"service_version")

            resp_status = parse_status(resp_obj.status)

            if int(resp_status[0]) >= 400:
                span.error_occurred = True
            resp = _original_falcon_api(this, env, start_response)
            span.tag(Tag(key=tags.HttpStatus, val=str(resp_status[0]), overridable=True))

        return resp

    def _fast_handle_exception(this: API, req, resp, ex, params):
        if ex is not None:
            entry_span = get_context().active_span()
            if entry_span is not None and type(entry_span) is not NoopSpan:
                entry_span.raised()

        return _original_falcon_handle_exception(this, req, resp, ex, params)

    def set_code(readerList, request, code_type="tenant_code"):
        """
        根据配置从不同渠道（cookie\header\querystring\env）获取tenant_code、user_code、env_code等数据
        :param reader:
        :param request:
        :param code_type:
        :return:
        """
        code = ""
        if not readerList:
            return code

        for reader in readerList:

            reader_type = reader.get("ReaderType")

            if reader_type == ReaderType.Cookie and request.cookies:
                    code = request.cookies.get(reader.get("ReaderKey"), "")
            elif reader_type == ReaderType.RequestHeader:
                # 下划线转中横杠
                code = ""
                if '_' in reader.get("ReaderKey"):
                    reader["ReaderKey"] = reader.get("ReaderKey").replace('_', '-')
                if reader.get("ReaderKey") in request.headers:
                    code = request.headers[reader.get("ReaderKey")]
                elif reader.get("ReaderKey").upper() in request.headers:
                    code = request.headers[reader.get("ReaderKey").upper()]
                elif reader.get("ReaderKey").capitalize() in request.headers:
                    code = request.headers[reader.get("ReaderKey").capitalize()]
            elif reader_type == ReaderType.QueryString:
                # import json
                # raw = request.stream.read()
                # raw_json = request.bounded_stream.read()
                # from urllib import parse
                # params = parse.parse_qsl(raw_json)
                code = request.params.get(reader.get("ReaderKey"), "")
            elif reader_type == ReaderType.Environment:
                import os

                code = os.getenv(reader.get("ReaderKey"), "")
            else:
                code = ""

            # functions.log("code_type:%s,reader_type:%s,reader_key:%s,code:%s",code_type,reader.get("ReaderType"),reader.get("ReaderKey"),code)

            if code != "":
                break

        # 设置用户自定义字段值
        if code_type == "tenant_code" and not config.custom_tenant_code:
            config.custom_tenant_code = code
        elif code_type == "user_code" and not config.custom_user_code:
            config.custom_user_code = code
        elif code_type == "service_version":
            config.custom_service_version = code

        return code

    API.__call__ = _fast_falcon_api
    API._handle_exception = _fast_handle_exception


def get_headers(env):
    """
    获取headers数据
    :param dict env: 环境变量
    :return:
    """
    headers = {}
    wsgi_content_headers = frozenset(["CONTENT_TYPE", "CONTENT_LENGTH"])

    for name, value in env.items():

        if name.startswith("HTTP_"):
            header_key = name[5:].replace("_", "-")
            if header_key == "FAST-TRACKER":
                value_arr = value.split("-")
                if len(value_arr) > 0:
                    headers[header_key] = lang.b64decode(value_arr[0])
                    if len(value_arr) >= 2:
                        headers["parent_id"] = lang.b64decode(value_arr[1])
            else:
                headers[header_key] = value
        elif name in wsgi_content_headers:
            headers[name.replace("_", "-")] = value

    return headers


def parse_status(status_str):
    """
    解析falcon的response的status数据，其结构类似：200 OK,参考：python3.8/site-packages/falcon/status_codes.py
    :param status_str:
    :return:
    """
    return status_str.split(" ") if status_str else [404, "status is empty"]



#!/usr/local/bin python3
# -*- coding: utf-8 -*-

"""
    created by FAST-DEV 2021/4/9
"""
import time
import sys
import os
from fast_tracker import config


def lower_case_name(text):
    """
    将驼峰命名转为小写下划线命名
    :param text:
    :return:
    """
    lst = []
    for index, char in enumerate(text):
        if char.isupper() and index != 0:
            lst.append("_")
        if char != ".":
            lst.append(char)

    return "".join(lst).lower()


def log(text, *args, level=1):
    """
    log日志标准输出
    :param text: 日志内容（含替换符）
    :param args: 替换符对应的数据
    :param int level: 日志级别, 数字越高，打印信息越多，为0表示不打印日志。暂定级别有：0，1，2，3，4 可对应为logging的日志级别
    :return:
    """
    if config.tracker_logging.get("Level"):
        text = text % args
        timestamp = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())
        sys.stdout.write("FAST: %s - %s\n" % (timestamp, text))
        sys.stdout.flush()


def getUri(text):
    if "http://" in text:
        index = text.find("//")
        domain = text[index + 2:]
        index = domain.find("/")
        return domain[index:]
    else:
        index = text.find("/")
    return text[index:]



#!/usr/local/bin python3
# -*- coding: utf-8 -*-

"""
    created by FAST-DEV 2021/4/6
    modified by cunlian 2021/12/02
"""
import inspect
import os
import re
import uuid
import sys
from typing import TYPE_CHECKING

from fast_tracker.utils import functions

if TYPE_CHECKING:
    from typing import List

import fast_tracker

RE_IGNORE_PATH = os.getenv("FastTracker_ReIgnorePath") or re.compile("^$")  # type: re.Pattern

service_instance = os.getenv("FastTracker_Instance") or str(uuid.uuid1()).replace("-", "")  # type: str
protocol = (os.getenv("FastTracker_Protocol") or "udp").lower()  # type: str
ignore_suffix = (
        os.getenv("FastTracker_IgnoreSuffix") or ".jpg,.jpeg,.js,.css,.png,.bmp,.gif,.ico,.mp3," ".mp4,.html,.svg "
)  # type: str
correlation_element_max_number = int(os.getenv("FastTracker_CorrelationElementMaxNumber") or "3")  # type: int
correlation_value_max_length = int(os.getenv("FastTracker_CorrelationValueMaxLength") or "128")  # type: int
trace_ignore_path = os.getenv("FastTracker_IgnorePath") or ""  # type: str

############################# fast_tracker默认配置信息开始 ###############################
# 以下配置位3.0版本的探针配置
tracker_version = fast_tracker.version

# agent全局开关，默认false
enable = False  # type: bool

# Debug开关
debug = False

# debug level
debug_level = 0

# 日志等级
logging_level = "ERROR"  # type: str

# 探针日志
tracker_logging = {"Level": "ERROR", "FilePath": "logs/fast-tracker/{Date}.log"}

# 服务名称
service_name = ""

# 服务版本
service_version = ""

# 服务版本codereader，优先环境变量，然后是cookie
service_version_reader = []

# 租户编码
tenant_code = ""  # type: str

# 租户codereader，优先环境变量，然后是cookie
tenant_code_reader = []

# 用户编码
user_code = ""  # type: str

# 用户编码codereader，优先环境变量，然后是cookie
user_code_reader = []

# 头部信息配置
carrier_header = {"TrackerName": "fast-tracker", "TraceIdName": "x-fast-trace-id"}

collector_address = "127.0.0.1:5140"
# 缓冲区大小 默认1
buffer_size = 1
# 上报超时时间 默认1（秒）
socket_timeout = 1

# 采集组件配置
collector_layer = {
    "HTTP": {
        "Enable": True,
        "CollectBodyContent": False,
        "CollectQueryString": False
    },
    "DB": {
        "Enable": True,
        "CollectParams": False
    },
    "MQ": {
        "Enable": True
    },
    "Cache": {
        "Enable": True
    },
    "RPC": {
        "Enable": True
    },
    "Log": {
        "Enable": True,
        "Level": "ERROR"
    },
    "Local": {
        "Enable": True
    },
    "Function": {
        "Enable": False,
        "ExcludeGetAndSetMethod": True,
        "ScanPackages": [],
        "IgnoreClasses": [],
        "IgnoreMethods": {}
    }
}
# 过滤设置
filter = {
    "IgnoreEntryPaths": [],
    "IgnoreEntryFiles": []
}

# 上报设置
transport = {
    "Forward": {
        "Endpoint": "udp://127.0.0.1:5140",
        "Timeout": 3,
        "Format": "json|msgpack"
    },
    "QueueSize": 1572864,
    "BatchSize": 1000,
    "Interval": 1000,
    "MaxReportByte": 32768
}

# 兼容原Event模块，目前一般只有Components
event = {"Components": {"SqlClient": True, "HttpClient": True}}
disable_plugins = []  # type: List[str]

############################# fast_tracker默认配置信息结束 ###############################
# 用户自定义user_code、tenant_code、env_code字段值
custom_user_code = ""
custom_tenant_code = ""
custom_env_code = ""
custom_service_version = ""


def init(
        service: str = None, instance: str = None, collector: str = None, protocol_type: str = "udp", token: str = None
):
    # 先根据环境变量 FastTracker_ConfigPath 加载配置文件并赋值
    config_file = os.environ.get("FastTracker.ConfigPath", None)
    print("FastTracker.ConfigPath is %s " % config_file)
    from fast_tracker.fast_tracker_configer import FastTrackerConfiger

    if config_file:
        FastTrackerConfiger.load_configuration(config_file)
    else:
        print("FAST: 探针配置文件未指定")

    # 再根据环境变量加载配置
    FastTrackerConfiger.set_config_by_env()

    # 再自定义设置的配置
    global service_name
    service_name = service or service_name

    global service_instance
    service_instance = instance or service_instance

    global collector_address
    collector_address = collector or collector_address

    global protocol
    protocol = protocol_type or protocol


def set_enable(cus_enable: bool = False):
    """
    设置enable值
    :param cus_enable:
    :return:
    """
    global enable
    enable = cus_enable


def set_debug(cus_debug: bool = False):
    """
    设置 debug 值
    :param cus_debug:
    :return:
    """
    global debug
    debug = cus_debug


def set_debug_level(cus_debug_level: int = 0):
    """
    设置 debug_level 值
    :param cus_debug_level:
    :return:
    """
    global debug_level
    debug_level = cus_debug_level


# def set_env_code(cus_env_code: str = ""):
#     """
#     设置env_code值
#     :param cus_env_code:
#     :return:
#     """
#     global env_code
#     env_code = cus_env_code


def set_tenant_code(cus_tenant_code: str = ""):
    """
    设置tenant_code值
    :param cus_tenant_code:
    :return:
    """
    global tenant_code
    tenant_code = cus_tenant_code


def set_user_code(cus_user_code: str = ""):
    """
    设置user_code值
    :param cus_user_code:
    :return:
    """
    global user_code
    user_code = cus_user_code


# def set_product_code(cus_product_code: str = ""):
#     """
#     设置 product_code 值
#     :param cus_product_code:
#     :return:
#     """
#     global product_code
#     product_code = cus_product_code


# def set_app_code(cus_app_code: str = ""):
#     """
#     设置 app_code 值
#     :param cus_app_code:
#     :return:
#     """
#     global app_code
#     app_code = cus_app_code


def set_service_name(cus_service_name: str = ""):
    """
    设置 service_name 值
    :param cus_service_name:
    :return:
    """
    global service_name
    service_name = cus_service_name


def set_service_version(cus_service_version: str = ""):
    """
    设置 service_version 值
    :param cus_service_version:
    :return:
    """
    global service_version
    service_version = cus_service_version
def get_service_version():
    return custom_service_version if custom_service_version else service_version

def set_socket_path(cus_collector_address: str = ""):
    """
    设置 collector_address 值
    :param cus_collector_address:
    :return:
    """
    global collector_address
    collector_address = cus_collector_address


def set_buffer_size(cus_buffer_size: int = 1):
    """
    设置 buffer_size 值
    :param cus_buffer_size:
    :return:
    """
    global buffer_size
    buffer_size = cus_buffer_size


def set_socket_timeout(cus_socket_timeout: int = 1):
    """
    设置 socket_timeout 值
    :param cus_socket_timeout:
    :return:
    """
    global socket_timeout
    socket_timeout = cus_socket_timeout


def set_event(cus_event: dict):
    """
    设置 event 值
    目前只有components的数据
    :param cus_event:
    :return:
    """
    global event
    if not cus_event:
        return False
    event = cus_event
    if event.get("Components"):
        cus_disable_plugins = []
        component = event.get("Components")
        if component.get("SqlClient") in ("false", "False", "0", False, 0, -1):
            cus_disable_plugins.extend(["fast_pymysql", "fast_pymongo", "fast_elasticsearch"])
        if component.get("HttpClient") in ("false", "False", "0", False, 0, -1):
            cus_disable_plugins.extend(
                [
                    "fast_django",
                    "fast_falcon",
                    "fast_flask",
                    "fast_http_server",
                    "fast_requests",
                    "fast_tornado",
                    "fast_urllib3",
                    "fast_urllib_request",
                ]
            )
        if component.get("CustomEvent") in ("false", "False", "0", False, 0, -1):
            cus_disable_plugins.append("custom_event")
        if component.get("Logging") in ("false", "False", "0", False, 0, -1):
            cus_disable_plugins.append("fast_log")

        if cus_disable_plugins:
            global disable_plugins
            disable_plugins = cus_disable_plugins


def set_tenant_code_reader(str={}):
    """
    设置 tenant_code_reader 值
    :param str:
    :return:
    """
    # global tenant_code_reader
    for key in str:
        if key.get("ReaderType") and key.get("ReaderKey"):

            tenant_code_reader.append(key)


def set_user_code_reader(str={}):
    """
    设置 user_code_reader 值
    :param str:
    :return:
    """
    global user_code_reader
    for key in str:
        if key.get("ReaderType") and key.get("ReaderKey"):
            user_code_reader.append(key)


def set_service_version_reader(str={}):
    """
    设置 service_version_reader 值
    :param str:
    :return:
    """
    global service_version_reader
    for key in str:
        if key.get("ReaderType") and key.get("ReaderKey"):
            functions.log("ReaderType:%s,ReaderKey:%s",key.get("ReaderType"),key.get("ReaderKey"))
            service_version_reader.append(key)


def set_carrier_header(**kwargs):
    """
    设置 carrier_header 值
    :param kwargs:
    :return:
    """
    global carrier_header
    if kwargs.get("TrackerName"):
        carrier_header["TrackerName"] = kwargs.get("TrackerName")

    if kwargs.get("TraceIdName"):
        carrier_header["TraceIdName"] = kwargs.get("TraceIdName")


def set_trace_id_name(trace_id_name: str):
    """
    设置trace_id_name
    :param str trace_id_name:
    :return:
    """
    global carrier_header
    if not trace_id_name:
        return False
    carrier_header["TraceIdName"] = trace_id_name


def get_trace_id_name():
    """
    获取trace_id_name
    :return str
    """
    global carrier_header
    return carrier_header.get("TraceIdName")


def set_tracker_name(tracker_name):
    """
    设置 tracker_name
    :param tracker_name:
    :return:
    """
    global carrier_header
    if not tracker_name:
        return False
    carrier_header["TrackerName"] = tracker_name


def get_tracker_name():
    """
    获取 tracker_name
    :return str
    """
    global carrier_header
    return carrier_header.get("TrackerName")


def set_logging(**kwargs):
    """
    设置 tracker_logging
    :param kwargs:
    :return:
    """
    global tracker_logging
    global logging_level
    if not kwargs.get("Level"):
        return False
    tracker_logging["Level"] = kwargs.get("Level").upper()
    logging_level = tracker_logging["Level"]
    if kwargs.get("FilePath"):
        tracker_logging["FilePath"] = kwargs.get("FilePath")


def set_collectLayer_http(**kwargs):
    if kwargs.get("Enable"):
        collector_layer["HTTP"]["Enable"] = kwargs.get("Enable")
    if kwargs.get("CollectBodyContent"):
        collector_layer["HTTP"]["CollectBodyContent"] = kwargs.get("CollectBodyContent")
    if kwargs.get("CollectQueryString"):
        collector_layer["HTTP"]["CollectQueryString"] = kwargs.get("CollectQueryString")


def set_collectLayer_db(**kwargs):
    if kwargs.get("Enable"):
        collector_layer["HTTP"]["Enable"] = kwargs.get("Enable")
    if kwargs.get("CollectParams"):
        collector_layer["HTTP"]["CollectParams"] = kwargs.get("CollectParams")


def set_collectLayer_mq(**kwargs):
    if kwargs.get("Enable"):
        collector_layer["HTTP"]["Enable"] = kwargs.get("Enable")


def set_collectLayer_cache(**kwargs):
    if kwargs.get("Enable"):
        collector_layer["HTTP"]["Enable"] = kwargs.get("Enable")


def set_collectLayer_rpc(**kwargs):
    if kwargs.get("Enable"):
        collector_layer["HTTP"]["Enable"] = kwargs.get("Enable")


def set_collectLayer_log(**kwargs):
    if kwargs.get("Enable"):
        collector_layer["HTTP"]["Enable"] = kwargs.get("Enable")
    if kwargs.get("Level"):
        collector_layer["HTTP"]["Level"] = kwargs.get("Level")


def set_collectLayer_local(**kwargs):
    if kwargs.get("Enable"):
        collector_layer["HTTP"]["Enable"] = kwargs.get("Enable")


def set_collectLayer_function(**kwargs):
    if kwargs.get("Enable"):
        collector_layer["Function"]["Enable"] = kwargs.get("Enable")
    if kwargs.get("ExcludeGetAndSetMethod"):
        collector_layer["Function"]["ExcludeGetAndSetMethod"] = kwargs.get("ExcludeGetAndSetMethod")
    if kwargs.get("ScanPackages"):
        collector_layer["Function"]["ScanPackages"] = kwargs.get("ScanPackages")
    if kwargs.get("IgnoreClasses"):
        collector_layer["Function"]["IgnoreClasses"] = kwargs.get("IgnoreClasses")
    if kwargs.get("IgnoreMethods"):
        collector_layer["Function"]["IgnoreMethods"] = kwargs.get("IgnoreMethods")


def set_collectLayer(**kwargs):
    """
    设置 collector_layer
    :param kwargs:
    :return:
    """
    global collector_layer
    if kwargs.get("HTTP"):
        set_collectLayer_http(**kwargs.get("HTTP"))
    if kwargs.get("DB"):
        set_collectLayer_db(**kwargs.get("DB"))
    if kwargs.get("MQ"):
        set_collectLayer_mq(**kwargs.get("MQ"))
    if kwargs.get("Cache"):
        set_collectLayer_cache(**kwargs.get("Cache"))
    if kwargs.get("RPC"):
        set_collectLayer_rpc(**kwargs.get("RPC"))
    if kwargs.get("Log"):
        set_collectLayer_log(**kwargs.get("Log"))
    if kwargs.get("Local"):
        set_collectLayer_local(**kwargs.get("Local"))
    if kwargs.get("Function"):
        set_collectLayer_function(**kwargs.get("Function"))


def set_filter(**kwargs):
    """
    设置 filter
    :param kwargs:
    :return:
    """
    global filter
    if kwargs.get("IgnoreEntryPaths"):
        filter["IgnoreEntryPaths"] = kwargs.get("IgnoreEntryPaths")

    if kwargs.get("IgnoreEntryFiles"):
        filter["IgnoreEntryFiles"] = kwargs.get("IgnoreEntryFiles")


def set_forward(**kwargs):
    global socket_timeout, collector_address
    if kwargs.get("Endpoint"):
        transport["Forward"]["Endpoint"] = kwargs.get("Endpoint")
        collector_address = kwargs.get("Endpoint")
    if kwargs.get("Timeout"):
        transport["Forward"]["Timeout"] = kwargs.get("Timeout")
        socket_timeout = kwargs.get("Timeout")
    if kwargs.get("Format"):
        transport["Forward"]["Format"] = kwargs.get("Format")


def set_transport(**kwargs):
    """
    设置 transport
    :param kwargs:
    :return:
    """
    global transport, buffer_size
    if kwargs.get("Forward"):
        set_forward(**kwargs.get("Forward"))
    if kwargs.get("QueueSize"):
        transport["QueueSize"] = kwargs.get("QueueSize")
        buffer_size = kwargs.get("QueueSize")
    if kwargs.get("BatchSize"):
        transport["BatchSize"] = kwargs.get("BatchSize")
    if kwargs.get("Interval"):
        transport["Interval"] = kwargs.get("Interval")
    if kwargs.get("MaxReportByte"):
        transport["MaxReportByte"] = kwargs.get("MaxReportByte")


def finalize():
    """
    通过忽略后缀的文件找到忽略的文件夹
    :return:
    """
    reesc = re.compile(r"([.*+?^=!:${}()|\[\]\\])")
    suffix = r"^.+(?:" + "|".join(reesc.sub(r"\\\1", s.strip()) for s in ignore_suffix.split(",")) + ")$"
    path = (
            "^(?:"
            + "|".join(  # replaces ","
        "(?:(?:[^/]+/)*[^/]+)?".join(  # replaces "**"
            "[^/]*".join(  # replaces "*"
                "[^/]".join(reesc.sub(r"\\\1", s) for s in p2.split("?")) for p2 in p1.split("*")  # replaces "?"
            )
            for p1 in p0.strip().split("**")
        )
        for p0 in trace_ignore_path.split(",")
    )
            + ")$"
    )

    global RE_IGNORE_PATH
    RE_IGNORE_PATH = re.compile("%s|%s" % (suffix, path))


def serialize():
    from fast_tracker import config

    return {
        key: value
        for key, value in config.__dict__.items()
        if not (
                key.startswith("_")
                or key == "TYPE_CHECKING"
                or key == "RE_IGNORE_PATH"
                or inspect.isfunction(value)
                or inspect.ismodule(value)
                or inspect.isbuiltin(value)
                or inspect.isclass(value)
        )
    }


def deserialize(data):
    from fast_tracker import config

    for key, value in data.items():
        if key in config.__dict__:
            config.__dict__[key] = value
    finalize()



#!/usr/local/bin python3
# -*- coding: utf-8 -*-

"""
    created by FAST-DEV 2021/4/12
"""

from fast_tracker import config, Layer, ComponentType, Log, LogItem
from fast_tracker.trace.context import get_context
from fast_tracker.trace.tags import Tag


class FastTracker:
    def __init__(self):
        self.span = None

    # @property
    # def product_code(self):
    #     """
    #     获取 product_code 值
    #     :return:
    #     """
    #     return config.product_code

    # @product_code.setter
    # def product_code(self, product_code: str = ""):
    #     """
    #     设置 product_code 值
    #     :param product_code:
    #     :return:
    #     """
    #     config.set_product_code(product_code)

    # @property
    # def app_code(self):
    #     """
    #     获取 app_code 值
    #     :return:
    #     """
    #     return config.app_code

    # @app_code.setter
    # def app_code(self, app_code: str = ""):
    #     """
    #     设置 app_code 值
    #     :param app_code:
    #     :return:
    #     """
    #     config.set_app_code(app_code)

    # @property
    # def env_code(self):
    #     """
    #     获取 env_code 值
    #     :return:
    #     """
    #     return config.custom_env_code if config.custom_env_code else config.env_code

    # @env_code.setter
    # def env_code(self, env_code: str = ""):
    #     """
    #     设置 env_code 值
    #     :param env_code:
    #     :return:
    #     """
    #     if env_code:
    #         config.custom_env_code = env_code

    @property
    def tenant_code(self):
        """
        获取 tenant_code 值
        :return:
        """
        return config.custom_tenant_code if config.custom_tenant_code else config.tenant_code

    @tenant_code.setter
    def tenant_code(self, tenant_code: str = ""):
        """
        设置 tenant_code 值
        :param tenant_code:
        :return:
        """
        return config.set_tenant_code(tenant_code)

    @property
    def user_code(self):
        """
        获取 user_code 值
        :return:
        """
        return config.custom_user_code if config.custom_user_code else config.user_code

    @user_code.setter
    def user_code(self, user_code: str = ""):
        """
        设置 user_code 值
        :param user_code:
        :return:
        """
        return config.set_user_code(user_code)

    @property
    def service_name(self):
        """
        获取 service_name 值
        :return:
        """
        return config.service_name

    @service_name.setter
    def service_name(self, service_name: str = ""):
        """
        设置 service_name 值
        :param service_name:
        :return:
        """
        config.set_service_name(service_name)

    @property
    def service_version(self):
        """
        获取 service_version 值
        :return:
        """
        return config.custom_service_version if config.custom_service_version else config.service_version

    @service_version.setter
    def service_version(self, service_version: str = ""):
        """
        设置 service_version 值
        :param service_version:
        :return:
        """
        config.set_service_version(service_version)

    @property
    def socket_path(self):
        """
        获取 socket_path 值
        :return:
        """
        return config.collector_address

    @socket_path.setter
    def socket_path(self, socket_path: str = ""):
        """
        设置 socket_path 值
        :param socket_path:
        :return:
        """
        config.set_socket_path(socket_path)

    @property
    def buffer_size(self):
        """
        获取 buffer_size 值
        :return:
        """
        return config.buffer_size

    @buffer_size.setter
    def buffer_size(self, buffer_size: int = 1):
        """
        设置 buffer_size 值
        :param buffer_size:
        :return:
        """
        config.set_buffer_size(buffer_size)

    @property
    def socket_timeout(self):
        """
        获取 socket_timeout 值
        :return:
        """
        return config.socket_timeout

    @socket_timeout.setter
    def socket_timeout(self, socket_timeout: int = 1):
        """
        设置 socket_timeout 值
        :param socket_timeout:
        :return:
        """
        config.set_socket_timeout(socket_timeout)

    @property
    def event(self):
        """
        获取 event 值
        :return:
        """
        return config.event

    @event.setter
    def event(self, event: dict):
        """
        设置 event 值
        :param event:
        :return:
        """
        config.set_event(event)

    @staticmethod
    def get_config():
        return {
            "ServiceName": config.service_name,
            "TrackerVersion": config.tracker_version,
            "ServiceInstance": config.service_instance,
            "Protocol": config.protocol,
            "LoggingLevel": config.logging_level,
            "IgnoreSuffix": config.ignore_suffix,
            "CorrelationElementMaxNumber": config.correlation_element_max_number,
            "CorrelationValueMaxLength": config.correlation_value_max_length,
            "TraceIgnorePath": config.trace_ignore_path,
            "Enable": config.enable,
            "Debug": config.debug,
            "DebugLevel": config.debug_level,
            # "EnvCode": config.env_code,
            "TenantCode": config.tenant_code,
            "UserCode": config.user_code,
            "SocketPath": config.collector_address,
            "BufferSize": config.buffer_size,
            "SocketTimeout": config.socket_timeout,
            "Event": config.event,
            "TenantCodeReader": config.tenant_code_reader,
            "UserCodeReader": config.user_code_reader,
            # "EnvCodeReader": config.env_code_reader,
            "CarrierHeader": config.carrier_header,
        }

    def begin_span(self, operation: str = ""):
        """
        log模块开始 类似数据库事务的begin_trasaction
        :param op:
        :return:
        """
        try:
            if "custom_event" in config.disable_plugins:
                raise RuntimeError("已在配置Event-Components-CustomEvent中关闭此项")
            if not operation:
                raise RuntimeError("operation参数必填")
            context = get_context()
            self.span = context.new_local_span(op="execute")
            self.span.layer = Layer.Local
            self.span.component = ComponentType.CustomEvent
            self.span.op = operation
            self.span.start()
            return self
        except Exception as e:
            raise e

    def set_component(self, name: str = "CustomEvent"):
        if "custom_event" in config.disable_plugins:
            raise RuntimeError("已在配置Event-Components-CustomEvent中关闭此项")
        if not self.span:
            raise RuntimeError("请先调用begin_span进行实例化")
        self.span.component = name

    def add_tag(self, key: str = "", val: str = ""):
        if "custom_event" in config.disable_plugins:
            raise RuntimeError("已在配置Event-Components-CustomEvent中关闭此项")
        if not self.span:
            raise RuntimeError("请先调用begin_span进行实例化")
        try:
            self.span.tag(Tag(key=key, val=val, overridable=True))
        except Exception as e:
            raise e

    def add_log(self, message: str = ""):
        if "custom_event" in config.disable_plugins:
            raise RuntimeError("已在配置Event-Components-CustomEvent中关闭此项")
        if not self.span:
            raise RuntimeError("请先调用begin_span进行实例化")
        self.span.logs.append(Log(items=[LogItem(key="Logging", val=self.log_covert(message, err_type=""))]))

    def error_occurred(self, error: Exception):
        if "custom_event" in config.disable_plugins:
            raise RuntimeError("已在配置Event-Components-CustomEvent中关闭此项")
        if not self.span:
            raise RuntimeError("请先调用begin_span进行实例化")
        self.span.log(error)

    def log_covert(self, msg: str = "", err_type: str = "DEBUG"):
        return {"err_msg": msg, "err_type": err_type if err_type else "", "err_trace": ""}

    def end_span(self):
        """
        log模块结束
        :return:
        """
        if "custom_event" in config.disable_plugins:
            raise RuntimeError("已在配置Event-Components-CustomEvent中关闭此项")
        if not self.span:
            raise RuntimeError("请先调用begin_span进行实例化")
        self.span.stop()

    def debug(self, msg: str = ""):
        self.__logging(msg, err_type="DEBUG")

    def info(self, msg: str = ""):
        self.__logging(msg, err_type="INFO")

    def warning(self, msg: str = ""):
        self.__logging(msg, err_type="WARNING")

    def warn(self, msg: str = ""):
        self.warning(msg)

    def error(self, msg: str = ""):
        self.__logging(msg, err_type="ERROR")

    def critical(self, msg: str = ""):
        self.__logging(msg, err_type="CRITICAL")

    def __logging(self, msg: str = "", err_type: str = ""):
        try:
            if "fast_log" in config.disable_plugins:
                raise RuntimeError("已在配置Event-Components-Logging中关闭此项")
            context = get_context()
            if not context:
                raise RuntimeError("探针未启动")
            with context.new_local_span(op=err_type) as span:
                span.layer = Layer.Local
                span.component = ComponentType.Log
                span.logs.append(Log(items=[LogItem(key="Logging", val=self.log_covert(msg, err_type="DEBUG"))]))
        except Exception as e:
            raise e



#!/usr/local/bin python3
# -*- coding: utf-8 -*-

"""
    created by FAST-DEV 2021/4/6
"""

from fast_tracker import Layer, ComponentType
from fast_tracker.trace import tags
from fast_tracker.trace.carrier import Carrier
from fast_tracker.trace.context import get_context
from fast_tracker.trace.tags import Tag


def install():
    from pika.channel import Channel

    _basic_publish = Channel.basic_publish
    __on_deliver = Channel._on_deliver
    Channel.basic_publish = _fast_basic_publish_func(_basic_publish)
    Channel._on_deliver = _fast__on_deliver_func(__on_deliver)


def _fast_basic_publish_func(_basic_publish):
    def _fast_basic_publish(this, exchange, routing_key, body, properties=None, mandatory=False):
        peer = "%s:%s" % (this.connection.params.host, this.connection.params.port)
        context = get_context()
        import pika

        with context.new_exit_span(
            op="RabbitMQ/Topic/" + exchange + "/Queue/" + routing_key + "/Producer" or "/", peer=peer
        ) as span:
            carrier = span.inject()
            span.layer = Layer.MQ
            span.component = ComponentType.RabbitmqProducer
            properties = pika.BasicProperties() if properties is None else properties

            if properties.headers is None:
                properties.headers = {}
            for item in carrier:
                properties.headers[item.key] = item.val

            res = _basic_publish(this, exchange, routing_key, body, properties=properties, mandatory=mandatory)
            span.tag(Tag(key=tags.MqType, val=ComponentType.Rabbitmq))
            span.tag(Tag(key=tags.MqBroker, val=peer))
            span.tag(Tag(key=tags.MqTopic, val=exchange))
            span.tag(Tag(key=tags.MqQueue, val=routing_key))

            return res

    return _fast_basic_publish


def _fast__on_deliver_func(__on_deliver):
    def _fast__on_deliver(this, method_frame, header_frame, body):
        peer = "%s:%s" % (this.connection.params.host, this.connection.params.port)
        context = get_context()
        exchange = method_frame.method.exchange
        routing_key = method_frame.method.routing_key
        carrier = Carrier()
        for item in carrier:
            if item.key in header_frame.properties.headers:
                item.val = header_frame.properties.headers[item.key]

        with context.new_entry_span(
            op="RabbitMQ/Topic/" + exchange + "/Queue/" + routing_key + "/Consumer" or "", carrier=carrier
        ) as span:
            span.layer = Layer.MQ
            span.component = ComponentType.RabbitmqConsumer
            __on_deliver(this, method_frame, header_frame, body)
            span.tag(Tag(key=tags.MqType, val=ComponentType.Rabbitmq))
            span.tag(Tag(key=tags.MqBroker, val=peer))
            span.tag(Tag(key=tags.MqTopic, val=exchange))
            span.tag(Tag(key=tags.MqQueue, val=routing_key))

    return _fast__on_deliver



#!/usr/local/bin python3
# -*- coding: utf-8 -*-

"""
    created by ty 2021/4/6
"""

import logging

from fast_tracker import config


def getLogger(name=None):
    logger = logging.getLogger(name)
    ch = logging.StreamHandler()
    formatter = logging.Formatter('%(name)s [%(threadName)s] [%(levelname)s] %(message)s')
    ch.setFormatter(formatter)
    # 设置流式处理器为日志处理器
    logger.addHandler(ch)
    # propagate
    # 如果此属性为true，则除了附加到此记录器的任何处理程序之外，记录到该记录器的事件还将传递到更高级别（祖先）记录器的处理程序。
    # 消息直接传递给祖先记录器的处理程序-既不考虑所讨论的祖先记录器的级别也没有过滤器。
    # 如果此结果为false，则不会将日志记录消息传递给祖先记录器的处理程序。
    logger.propagate = False

    return logger


logger = getLogger('fast_tracker')


def init():
    # CRITICAL ERROR WARNING  INFO    DEBUG    NOTSET
    logging.addLevelName(logging.CRITICAL + 10, 'OFF')
    if config.tracker_logging.get("Level"):
        logger.setLevel(logging.getLevelName(config.logging_level))



#!/usr/local/bin python3
# -*- coding: utf-8 -*-

"""
    created by FAST-DEV 2021/4/6
"""

from fast_tracker import Layer, ComponentType, config
from fast_tracker.trace import tags
from fast_tracker.trace.carrier import Carrier
from fast_tracker.trace.context import get_context
from fast_tracker.trace.tags import Tag
from fast_tracker.utils import functions
from fast_tracker.utils.reader_type import ReaderType

version_rule = {"name": "django", "rules": [">=2.0"]}


def install():
    from django.core.handlers.base import BaseHandler
    from django.core.handlers import exception

    _get_response = BaseHandler.get_response
    _handle_uncaught_exception = exception.handle_uncaught_exception

    def _fast_get_response(this, request):
        if request is None:
            resp = _get_response(this, request)
            return resp

        context = get_context()
        carrier = Carrier()
        for item in carrier:
            # Any HTTP headers in the request are converted to META keys by converting all characters to uppercase,
            # replacing any hyphens with underscores and adding an HTTP_ prefix to the name.
            # https://docs.djangoproject.com/en/3.0/ref/request-response/#django.http.HttpRequest.META
            fast_http_header_key = "HTTP_%s" % item.key.upper().replace("-", "_")
            if fast_http_header_key in request.META:
                item.val = request.META[fast_http_header_key]

        functions.log("django url:%s", str(request.path))
        with context.new_entry_span(op=request.path.split("?")[0], carrier=carrier) as span:
            span.layer = Layer.Http
            span.component = ComponentType.Django
            span.peer = "%s:%s" % (request.META.get("REMOTE_ADDR"), request.META.get("REMOTE_PORT") or "80")

            span.tag(Tag(key=tags.HttpMethod, val=request.method))
            http_url = request.build_absolute_uri().split("?")[0]
            span.op = functions.getUri(http_url)
            span.tag(Tag(key=tags.HttpUrl, val=http_url))
            span.tag(Tag(key=tags.HttpPath, val=request.META.get("PATH_INFO")))

            # tenant_code\user_code\env_code 根据配置动态获取并更新
            set_code(config.tenant_code_reader, request, "tenant_code")
            set_code(config.user_code_reader, request, "user_code")
            set_code(config.service_version_reader,request,"service_version")
            # set_code(config.env_code_reader, request, "env_code")

            # you can get request parameters by `request.GET` even though client are using POST or other methods
            if request.GET:
                span.tag(
                    Tag(key=tags.HttpParams, val=params_tostring(request.GET)[0:])
                )

            resp = _get_response(this, request)
            span.tag(Tag(key=tags.HttpStatus, val=resp.status_code, overridable=True))
            if resp.status_code >= 400:
                span.error_occurred = True
            functions.log("django span is: %r", span)
            return resp

    def _fast_handle_uncaught_exception(request, resolver, exc_info):
        if exc_info is not None:
            entry_span = get_context().active_span()
            if entry_span is not None:
                entry_span.raised()

        return _handle_uncaught_exception(request, resolver, exc_info)

    BaseHandler.get_response = _fast_get_response
    exception.handle_uncaught_exception = _fast_handle_uncaught_exception

    def set_code(readerList, request, type="tenant_code"):
        """
        根据配置从不同渠道（cookie\header\querystring\env）获取tenant_code、user_code、env_code等数据
        :param reader:
        :param request:
        :param type:
        :return:
        """
        code = ""
        if not readerList:
            return code

        for reader in readerList:
            reader_type = reader.get("ReaderType")
            reader_key = reader.get("ReaderKey")

            if reader_type == ReaderType.Cookie and request.cookies:
                code = request.COOKIES.get(reader_key, "")
            elif reader_type == ReaderType.RequestHeader:
                code = request.META[reader_key] if reader_key in request.headers else ""
            elif reader_type == ReaderType.QueryString:
                if request.method == "GET":
                    code = request.GET.get(reader_key, default="")
                elif request.method == "POST":
                    code = request.POST.get(reader_key, default="")
            elif reader_type == ReaderType.Environment:
                import os

                code = os.getenv(reader_key, "")
            else:
                code = ""
            if code != "":
                break

        if code:
            # 设置config的code，让其全局适用
            if type == "tenant_code":
                config.set_tenant_code(code)
            elif type == "user_code":
                config.set_user_code(code)
            elif type == "service_version":
                config.set_service_version(code)

        return code


def params_tostring(params):
    return "\n".join([k + "=[" + ",".join(params.getlist(k)) + "]" for k, _ in params.items()])



#!/usr/local/bin python3
# -*- coding: utf-8 -*-

"""
    created by FAST-DEV 2021/4/6
"""

from fast_tracker import Layer, ComponentType
from fast_tracker.trace import tags
from fast_tracker.trace.context import get_context
from fast_tracker.trace.tags import Tag


def install():
    from elasticsearch import Transport

    _perform_request = Transport.perform_request

    def _fast_perform_request(this: Transport, method, url, headers=None, params=None, body=None):
        context = get_context()
        peer = ",".join([host["host"] + ":" + str(host["port"]) for host in this.hosts])
        with context.new_exit_span(op="Elasticsearch/" + method + url, peer=peer) as span:
            span.layer = Layer.Database
            span.component = ComponentType.Elasticsearch
            res = _perform_request(this, method, url, headers=headers, params=params, body=body)

            span.tag(Tag(key=tags.DbType, val=ComponentType.Elasticsearch))
            # have no instance
            span.tag(Tag(key=tags.DbInstance, val=""))
            span.tag(Tag(key=tags.DbStatement, val="" if body is None else body))
            # have no parameters
            span.tag(Tag(key=tags.DbSqlParameters, val=""))

            return res

    Transport.perform_request = _fast_perform_request



#!/usr/local/bin python3
# -*- coding: utf-8 -*-

"""
    created by FAST-DEV 2021/4/6
"""
from inspect import iscoroutinefunction, isawaitable

from fast_tracker import Layer, ComponentType, config
from fast_tracker.trace import tags
from fast_tracker.trace.carrier import Carrier
from fast_tracker.trace.context import get_context
from fast_tracker.trace.tags import Tag
from fast_tracker.utils import functions
from fast_tracker.utils.reader_type import ReaderType


def install():
    # todo: 后续需要对此插件做版本控制，现在仅支持5.1.1版本， 6+没支持
    from tornado.web import RequestHandler

    old_execute = RequestHandler._execute
    old_log_exception = RequestHandler.log_exception
    RequestHandler._execute = _gen_fast_get_response_func(old_execute)

    def _fast_handler_uncaught_exception(self: RequestHandler, ty, value, tb, *args, **kwargs):
        if value is not None:
            entry_span = get_context().active_span()
            if entry_span is not None:
                entry_span.raised()

        return old_log_exception(self, ty, value, tb, *args, **kwargs)

    RequestHandler.log_exception = _fast_handler_uncaught_exception


def _gen_fast_get_response_func(old_execute):
    from tornado.gen import coroutine

    awaitable = iscoroutinefunction(old_execute)
    if awaitable:
        # Starting Tornado 6 RequestHandler._execute method is a standard Python coroutine (async/await)
        # In that case our method should be a coroutine function too
        async def _fast_get_response(self, *args, **kwargs):
            request = self.request
            context = get_context()
            carrier = Carrier()
            trace_id_name = config.get_trace_id_name()
            for item in carrier:
                if item.key.capitalize() in request.headers:
                    item.val = request.headers[item.key.capitalize()]
                if item.key.upper() in request.headers:
                    item.val = request.headers[item.key.upper()]
                if trace_id_name.capitalize() in request.headers and type(item) is Carrier:
                    item.set_frontend_trace_id(request.headers[trace_id_name.capitalize()])
                if trace_id_name.upper() in request.headers and type(item) is Carrier:
                    item.set_frontend_trace_id(request.headers[trace_id_name.upper()])
            functions.log("fast_tornado op:%s",request.path)
            with context.new_entry_span(op=request.path.split("?")[0], carrier=carrier) as span:
                span.layer = Layer.Http
                span.component = ComponentType.Tornado
                try:
                    # self.request.connection.context.address  get ('::1', 52857, 0, 0) sometimes got error
                    span.peer = self.request.host  # got like: localhost:8888
                except BaseException as e:
                    span.peer = self.request.headers.get("X-Real-IP") or \
                            self.request.headers.get("X-Forwarded-For") or \
                            self.request.remote_ip  # ::1
                    pass
                span.tag(Tag(key=tags.HttpMethod, val=request.method))
                span.tag(Tag(key=tags.HttpUrl, val="{}://{}{}".format(request.protocol, request.host, request.path)))
                # HttpPath 只需要获取PATH_INFO信息即可，不需要？后的内容
                span.tag(Tag(key=tags.HttpPath, val=(request.path).split("?")[0]))
                result = old_execute(self, *args, **kwargs)
                if isawaitable(result):
                    result = await result
                span.tag(Tag(key=tags.HttpStatus, val=self._status_code, overridable=True))
                if self._status_code >= 400:
                    span.error_occurred = True

                # tenant_code\user_code\env_code 根据配置动态获取并更新
                set_code(config.tenant_code_reader, request, "tenant_code")
                set_code(config.user_code_reader, request, "user_code")
                set_code(config.service_version_reader,request,"service_version")
                # set_code(config.env_code_reader, request, "env_code")
            return result

    else:

        @coroutine
        def _fast_get_response(self, *args, **kwargs):
            request = self.request
            context = get_context()
            carrier = Carrier()
            trace_id_name = config.get_trace_id_name()
            for item in carrier:
                if item.key.capitalize() in request.headers:
                    item.val = request.headers[item.key.capitalize()]
                if item.key.upper() in request.headers:
                    item.val = request.headers[item.key.upper()]
                if trace_id_name.capitalize() in request.headers and type(item) is Carrier:
                    item.set_frontend_trace_id(request.headers[trace_id_name.capitalize()])
                if trace_id_name.upper() in request.headers and type(item) is Carrier:
                    item.set_frontend_trace_id(request.headers[trace_id_name.upper()])
            functions.log("fast_tornado op:%s",request.path)
            with context.new_entry_span(op=request.path.split("?")[0], carrier=carrier) as span:
                span.layer = Layer.Http
                span.component = ComponentType.Tornado
                # v6+use this
                # peer = request.connection.stream.socket.getpeername()
                # peer = self.stream.socket.getpeername()
                # v5.1.1 use this
                try:
                    # self.request.connection.context.address  get ('::1', 52857, 0, 0) sometimes got error
                    span.peer = self.request.host
                except BaseException as e:
                    span.peer = self.request.headers.get("X-Real-IP") or \
                            self.request.headers.get("X-Forwarded-For") or \
                            self.request.remote_ip  # ::1
                    pass
                span.tag(Tag(key=tags.HttpMethod, val=request.method))
                span.tag(Tag(key=tags.HttpUrl, val="{}://{}{}".format(request.protocol, request.host, request.path)))
                result = yield from old_execute(self, *args, **kwargs)
                span.tag(Tag(key=tags.HttpStatus, val=self._status_code, overridable=True))
                if self._status_code >= 400:
                    span.error_occurred = True

                set_code(config.tenant_code_reader, request, "tenant_code")
                set_code(config.user_code_reader, request, "user_code")
                set_code(config.service_version_reader,request,"service_version")
                # set_code(config.env_code_reader, request, "env_code")

                functions.log("tornado span is: %r", span)
            return result

    return _fast_get_response


def set_code(readerList, request, code_type="tenant_code"):
    """
    根据配置从不同渠道（cookie\header\querystring\env）获取tenant_code、user_code、env_code等数据
    :param reader:
    :param request:
    :param code_type:
    :return:
    """
    code = ""
    if not readerList:
        return code

    for reader in readerList:

        reader_type = reader.get("ReaderType")
        reader_key = reader.get("ReaderKey")

        try:
            if reader_type == ReaderType.Cookie and request.cookies:
                # # tornado6.0.4 is ok
                # code = request.get_cookie(reader_key, "")
                # tornado5.1.1 is ok
                # code: Set-Cookie: env_c=gog <Morsel: env_c=gog>
                # request.cookies is A dictionary of Cookie.Morsel objects
                # tornado5.1.1 source code->tornado-httputil.py-HTTPServerRequest-cookies
                cookie = request.cookies.get(reader_key, "")
                if type(cookie) == str:
                    code = cookie
                else:
                    code = request.cookies.get(reader_key, "").value
                # code = request.cookies.get(reader_key, "").value
            elif reader_type == ReaderType.RequestHeader:
                code = request.headers[reader_key] if reader_key in request.headers else ""
            elif reader_type == ReaderType.QueryString:
                # data = json.loads(request.body.decode('utf-8'))
                # # tornado6.0.4 is ok
                # code = request.get_argument(reader_key, default="") if request.arguments.has_key(reader_key) else ""
                # tornado5.1.1
                arg = request.arguments.get(reader_key)
                code = arg[0].decode("utf-8") if type(arg) == list else ""
            elif reader_type == ReaderType.Environment:
                import os

                code = os.getenv(reader.get("ReaderKey"), "")
            else:
                code = ""

            if code != "":
                break
        except BaseException as e:
            functions.log("tornado 获取code错误，错误信息: %r", str(e))

    # 设置用户自定义字段值
    if code_type == "tenant_code" and not config.custom_tenant_code:
        config.custom_tenant_code = code
    elif code_type == "user_code" and not config.custom_user_code:
        config.custom_user_code = code
    elif code_type == "service_version" :
        config.custom_service_version = code
    return code



#!/usr/local/bin python3
# -*- coding: utf-8 -*-

"""
    created by FAST-DEV 2021/4/6
"""
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from fast_tracker.trace.context import SpanContext

from fast_tracker.trace import ID


class Snapshot:
    def __init__(
        self,
        segment_id: str = None,
        span_id: str = None,
        trace_id: ID = None,
        endpoint: str = None,
        correlation: dict = None,
    ):
        self.trace_id = trace_id  # type: ID
        self.segment_id = segment_id  # type: str
        self.span_id = span_id  # type: str
        self.endpoint = endpoint  # type: str
        self.correlation = correlation.copy()  # type: dict

    def is_from_current(self, context: "SpanContext"):
        return self.segment_id is not None and self.segment_id == context.capture().segment_id

    def is_valid(self):
        return self.trace_id is not None



#!/usr/local/bin python3
# -*- coding: utf-8 -*-

"""
    created by FAST-DEV 2021/4/6
"""

import uuid

from fast_tracker.utils.counter import AtomicCounter

_id = AtomicCounter()


class ID(object):
    def __init__(self, raw_id: str = None):
        self.value = raw_id or str(uuid.uuid1()).replace('-', '')

    def __str__(self):
        return self.value



#!/usr/local/bin python3
# -*- coding: utf-8 -*-

"""
    created by FAST-DEV 2021/4/6
"""
import json
import os
import socket
from fast_tracker import config
from fast_tracker.fast_tracker import FastTracker
from fast_tracker.client import TraceSegmentReportService
from fast_tracker.report.report_struct import ReportStruct
from fast_tracker.utils import functions

DEFAULT_UDP_IP = "127.0.0.1"
DEFAULT_UDP_PORT = 5140


class UdpTraceSegmentReportService(TraceSegmentReportService):
    def __init__(self):
        self.udp_ip = DEFAULT_UDP_IP
        self.udp_port = DEFAULT_UDP_PORT
        if config.collector_address:
            self._init_ip_and_port(config.collector_address)
        self.buffer_size = config.transport.get("BatchSize")
        self.type_dict = {"Entry": 0, "Exit": 1, "Local": 2}
        self.layer_dict = {
            "local": 0,
            "database": 1,
            "rpc_framework": 2,
            "http": 3,
            "mq": 4,
            "cache": 5,
        }

    def _init_ip_and_port(self, collector):
        """
        初始化udp数据上报的ip和port信息
        :param str collector: 上报数据的ip和port信息合并在一起的字符串
        :return:
        """
        if not collector:
            functions.log("socket_path参数为空")
            return False

        connections = collector.split(":")
        if len(connections) == 2:
            self.udp_ip, self.udp_port = connections
            self.udp_port = int(self.udp_port)
        elif len(connections) == 3:
            self.udp_ip = connections[1].strip("//")
            self.udp_port = int(connections[2]) if len(connections) > 2 else DEFAULT_UDP_PORT
        else:
            functions.log("socket_path参数不合法")
            return False

    def report(self, generator):
        """
        udp 数据包发送
        :param generator:
        :return:
        """
        try:
            functions.log("udp socket start", level=4)
            udp_sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)  # Internet  # UDP
            functions.log("udp socket object success")
            for segment in generator:
                for span in segment.spans:
                    # functions.log("udp socket object will get data", level=4)
                    data = self._get_data(segment, span)
                    functions.log("udp socket ip: %s, port: %d", self.udp_ip, self.udp_port)
                    # functions.log("udp socket object get data", level=4)
                    res = udp_sock.sendto(data, (self.udp_ip, self.udp_port))
                    # res = udp_sock.sendto(data, ("fast-collector", self.udp_port))
                    # functions.log("sended udp socket ip: %s, port: %d", self.udp_ip, self.udp_port)
                    functions.log("sended message: %r, result is: %r", data, res)
                # segment上报完清除用户自定义字段值，避免污染之后的上报
                config.custom_env_code = ""
                config.custom_tenant_code = ""
                config.custom_user_code = ""
        except BaseException as e:
            functions.log("udp report error: %r", e)


    def _get_data(self, segment, span):
        """
        udp 数据包组装
        :param segment:
        :param span:
        :return:
        """
        ft = FastTracker()
        report_struct = ReportStruct()
        # report_struct.product_code = config.product_code
        # report_struct.app_code = config.app_code
        report_struct.service_name = config.service_name
        # report_struct.env_code = ft.env_code
        report_struct.tenant_code = ft.tenant_code
        report_struct.user_code = ft.user_code
        report_struct.trace_id = str(segment.related_traces[0])
        report_struct.parent_span_id = span.pid
        report_struct.span_id = span.sid
        report_struct.peer = span.peer
        report_struct.span_type = self.type_dict.get(span.kind.name, 3)
        report_struct.span_layer = self.layer_dict.get(str(span.layer.name).lower(), 10)
        report_struct.start_time = span.start_time
        report_struct.end_time = span.end_time
        report_struct.duration = 1 if span.end_time - span.start_time == 0 else span.end_time - span.start_time
        report_struct.component = span.component
        report_struct.operation_name = span.op
        report_struct.is_error = span.error_occurred
        report_struct.tags = self.convert_tags(span.tags)
        report_struct.logs = self.convert_logs(span.logs)
        report_struct.service_instance = config.service_name + "|" + str(os.getpid())+"|"+socket.gethostname()
        report_struct.service_version = config.get_service_version()
        report_struct.tracker_version = config.tracker_version
        item = ReportStruct.convert(report_struct)
        #functions.log("report_struct: %r", item, level=2)
        data = json.dumps(item).encode("utf-8")
        return data

    def convert_tags(self, tags_data):
        """
        组装上报的tags数据结构
        :param tags_data:
        :return:
        """
        data = {}
        if not tags_data:
            return data
        for tag in tags_data:
            data[tag.key] = str(tag.val)

        return data

    def convert_logs(self, logs_data):
        """
        组装上报的logs数据结构
        :param logs_data:
        :return:
        """
        data = []
        if not logs_data:
            return data
        for log in logs_data:
            for item in log.items:
                if type(item.val) == str:
                    alpha = {
                        "Timestamp": int(log.timestamp * 1000),
                        "Data": {
                            "event": item.key,
                            "message": item.val,
                            "error_kind": "Error",
                            # "stack": "",
                        },
                    }
                else:
                    temp = {"event": item.key, "message": item.val.get("err_msg")}
                    if item.val.get("err_type"):
                        temp["error_kind"] = item.val.get("err_type")
                    if item.val.get("err_trace"):
                        temp["stack"] = item.val.get("err_trace")
                    alpha = {
                        "Timestamp": int(log.timestamp * 1000),
                        "Data": temp,
                    }
                data.append(alpha)

        return data



#!/usr/local/bin python3
# -*- coding: utf-8 -*-

"""
    created by FAST-DEV 2021/4/9
"""
import json
import os
import sys

from fast_tracker import config
from fast_tracker.utils import functions


class FastTrackerConfiger:
    @staticmethod
    def _default_config_keys():
        return [
            "Enable",
            "Logging",
            "ServiceName",
            "ServiceVersion",
            "TenantCode",
            "UserCode",
            "CollectLayer",
            "Filter",
            "Transport",
            "Logging",
            "TenantCodeReader",
            "UserCodeReader",
            "CarrierHeader",
            "ServiceVersionReader"
        ]

    @staticmethod
    def load_configuration(config_file=None):
        """
        :param config_file: 配置文件地址
        :return:
        """
        if not config_file:
            functions.log("没有探针配置文件")
            return False
        # functions.log("探针配置文件: %r", config_file)

        try:
            with open(config_file, 'r') as fb:
                config_dict = json.load(fb)
                default_config_keys = FastTrackerConfiger._default_config_keys()

                if config_dict:
                    for config_key in config_dict.keys():
                        if config_key in default_config_keys:
                            func_name = "set_" + functions.lower_case_name(config_key)
                            if config_key == 'TenantCodeReader':
                                config.set_tenant_code_reader(config_dict.get(config_key))
                            elif config_key == "UserCodeReader":
                                config.set_user_code_reader(config_dict.get(config_key))
                            elif config_key == "ServiceVersionReader":
                                config.set_service_version_reader(config_dict.get(config_key))
                            elif config_key == "CarrierHeader":
                                config.set_carrier_header(**config_dict.get(config_key))
                            elif config_key == "Logging":
                                config.set_logging(**config_dict.get(config_key))
                            elif config_key == "CollectLayer":
                                config.set_collectLayer(**config_dict.get(config_key))
                            elif config_key == "Filter":
                                config.set_filter(**config_dict.get(config_key))
                            elif config_key == "Transport":
                                config.set_transport(**config_dict.get(config_key))
                            else:
                                getattr(config, func_name)(config_dict.get(config_key))

        except Exception as e:
            functions.log("python探针初始化失败！json格式配置文件格式不合法(不要有注释),解析失败,文件: %s, 错误信息：%s", config_file, str(e))

    @staticmethod
    def set_config_by_env():
        """
        通过环境变量设置配置值
        :return:
        """
        try:
            env_dict = os.environ
            for key, val in env_dict.items():
                if key.startswith("FastTracker."):
                    if key.startswith("FastTracker.TenantCodeReader"):
                        config.tenant_code_reader.clear()
                        config.set_tenant_code_reader(json.loads(val))
                    elif key.startswith("FastTracker.UserCodeReader"):
                        config.user_code_reader.clear()
                        config.set_user_code_reader(json.loads(val))
                    elif key.startswith("FastTracker.ServiceVersionReader"):
                        config.service_version_reader.clear()
                        config.set_service_version_reader(json.loads(val))
                    elif key.startswith("FastTracker.CarrierHeader."):
                        args = {key[26:]: val}
                        config.set_carrier_header(**args)
                    elif key.startswith("FastTracker.Logging."):
                        args = {key[20:]: val}
                        config.set_logging(**args)
                    elif key.startswith("FastTracker.Filter."):
                        args = {key[19:]: val}
                        config.set_logging(**args)
                    elif key.startswith("FastTracker.Transport."):
                        if key.startswith("FastTracker.Transport.Forward."):
                            args = {key[30:]: val}
                            config.set_forward(**args)
                        else:
                            args = {key[22:]: val}
                            config.set_transport(**args)
                    elif key.startswith("FastTracker.CollectLayer."):
                        if key.startswith("FastTracker.CollectLayer.HTTP."):
                            args = {key[30:]: val}
                            config.set_collectLayer_http(**args)
                        elif key.startswith("FastTracker.CollectLayer.DB."):
                            args = {key[28:]: val}
                            config.set_collectLayer_db(**args)
                        elif key.startswith("FastTracker.CollectLayer.MQ."):
                            args = {key[28:]: val}
                            config.set_collectLayer_mq(**args)
                        elif key.startswith("FastTracker.CollectLayer.Cache."):
                            args = {key[31:]: val}
                            config.set_collectLayer_cache(**args)
                        elif key.startswith("FastTracker.CollectLayer.Log."):
                            args = {key[29:]: val}
                            config.set_collectLayer_log(**args)
                        elif key.startswith("FastTracker.CollectLayer.Local."):
                            args = {key[31:]: val}
                            config.set_collectLayer_local(**args)
                        elif key.startswith("FastTracker.CollectLayer.Function."):
                            args = {key[34:]: val}
                            config.set_collectLayer_function(**args)
                    else:
                        config_name = functions.lower_case_name(key[12:])
                        setattr(config, config_name, val)
                        # functions.log("config_name:%s,debug_level:%s,enable:%s,debug:%s", config_name, str(config.debug_level),str(config.enable),str(config.debug))
        except Exception as e:
            functions.log("python探针环境变量读取错误信息：%s", str(e))



# -*- coding: utf-8 -*-

from __future__ import print_function

from fast_tracker.admin import command


@command('run-python', '...',
         """使用提供的参数执行Python解释器,启动时会强行初始化代理. 环境变量FastTrackerConfigPath可以提供代理配置文件的路径,日志文件的详情由FAST_LOG提供""")
def run_python(args):
    import os
    import sys
    import time

    startup_debug = os.environ.get('FAST_STARTUP_DEBUG',
                                   'off').lower() in ('on', 'true', '1')

    def log_message(text, *args):
        if startup_debug:
            text = text % args
            timestamp = time.strftime('%Y-%m-%d %H:%M:%S', time.localtime())
            print('FAST: %s (%d) - %s' % (timestamp, os.getpid(), text))

    log_message('FAST Admin Script (%s)', __file__)

    log_message('working_directory = %r', os.getcwd())
    log_message('current_command = %r', sys.argv)

    log_message('sys.prefix = %r', os.path.normpath(sys.prefix))

    try:
        log_message('sys.real_prefix = %r', sys.real_prefix)
    except AttributeError:
        pass

    log_message('sys.version_info = %r', sys.version_info)
    log_message('sys.executable = %r', sys.executable)
    log_message('sys.flags = %r', sys.flags)
    log_message('sys.path = %r', sys.path)

    for name in sorted(os.environ.keys()):
        if name.startswith('FAST_') or name.startswith('PYTHON'):
            log_message('%s = %r', name, os.environ.get(name))

    from fast_tracker import __file__ as root_directory

    root_directory = os.path.dirname(root_directory)
    boot_directory = os.path.join(root_directory, 'bootstrap')

    log_message('root_directory = %r', root_directory)
    log_message('boot_directory = %r', boot_directory)

    python_path = boot_directory

    if 'PYTHONPATH' in os.environ:
        path = os.environ['PYTHONPATH'].split(os.path.pathsep)
        if not boot_directory in path:
            python_path = "%s%s%s" % (boot_directory, os.path.pathsep,
                                      os.environ['PYTHONPATH'])

    os.environ['PYTHONPATH'] = python_path

    os.environ['FAST_ADMIN_COMMAND'] = repr(sys.argv)

    os.environ['FAST_PYTHON_PREFIX'] = os.path.realpath(
        os.path.normpath(sys.prefix))
    os.environ['FAST_PYTHON_VERSION'] = '.'.join(
        map(str, sys.version_info[:2]))

    bin_directory = os.path.dirname(sys.argv[0])

    if bin_directory:
        python_exe = os.path.basename(sys.executable)
        python_exe_path = os.path.join(bin_directory, python_exe)
        if (not os.path.exists(python_exe_path) or
                not os.access(python_exe_path, os.X_OK)):
            python_exe_path = sys.executable
    else:
        python_exe_path = sys.executable

    log_message('python_exe_path = %r', python_exe_path)
    log_message('execl_arguments = %r', [python_exe_path, python_exe_path] + args)

    os.execl(python_exe_path, python_exe_path, *args)



#!/usr/local/bin python3
# -*- coding: utf-8 -*-

"""
    created by FAST-DEV 2021/4/6
"""

import time
from typing import List, TYPE_CHECKING

from fast_tracker import config

from fast_tracker.trace import ID
from fast_tracker.utils.lang import tostring

if TYPE_CHECKING:
    from fast_tracker.trace.carrier import Carrier
    from fast_tracker.trace.span import Span
    from fast_tracker.trace.snapshot import Snapshot


class SegmentRef(object):
    def __init__(self, carrier: "Carrier", ref_type: str = "CrossProcess"):
        self.ref_type = ref_type  # type: str
        self.trace_id = carrier.trace_id  # type: str
        self.segment_id = carrier.segment_id  # type: str
        self.span_id = carrier.span_id if carrier.span_id else ""  # type: str
        self.service = carrier.service  # type: str
        self.service_instance = carrier.service_instance  # type: str
        self.endpoint = carrier.endpoint  # type: str
        self.client_address = carrier.client_address  # type: str

    def __eq__(self, other):
        if not isinstance(other, SegmentRef):
            raise NotImplementedError
        return (
            self.ref_type == other.ref_type
            and self.trace_id == other.trace_id
            and self.segment_id == other.segment_id
            and self.span_id == other.span_id
            and self.service == other.service
            and self.service_instance == other.service_instance
            and self.endpoint == other.endpoint
            and self.client_address == other.client_address
        )

    @classmethod
    def build_ref(cls, snapshot: "Snapshot"):
        from fast_tracker.trace.carrier import Carrier

        carrier = Carrier()
        carrier.trace_id = str(snapshot.trace_id)
        carrier.segment_id = str(snapshot.segment_id)
        carrier.endpoint = snapshot.endpoint
        carrier.span_id = snapshot.span_id
        carrier.service = config.service_name
        carrier.service_instance = config.service_instance
        return SegmentRef(carrier, ref_type="CrossThread")


class _NewID(ID):
    pass


@tostring
class Segment(object):
    def __init__(self):
        self.segment_id = ID()  # type: ID
        self.spans = []  # type: List[Span]
        self.timestamp = int(time.time() * 1000)  # type: int
        self.related_traces = [_NewID()]  # type: List[ID]

    def archive(self, span: "Span"):
        self.spans.append(span)

    def relate(self, trace_id: ID):
        if isinstance(self.related_traces[0], _NewID):
            del self.related_traces[-1]
        self.related_traces.append(trace_id)



#!/usr/local/bin python3
# -*- coding: utf-8 -*-

"""
    created by FAST-DEV 2021/4/6
"""
import logging
from fast_tracker import Layer, ComponentType, Log, LogItem, config
from fast_tracker.trace.context import get_context
from fast_tracker.utils import functions

log_level = {"NOTSET": 0, "DEBUG": 10, "INFO": 20, "WARNING": 30, "ERROR": 40, "CRITICAL": 50}


def install():
    """
    python日志拦截器
    lib/python3.8/loggin/__init__.py中已定义日志等级:
    CRITICAL = 50
    FATAL = CRITICAL
    ERROR = 40
    WARNING = 30
    WARN = WARNING
    INFO = 20
    DEBUG = 10
    NOTSET = 0
    :return:
    """
    _handle = logging.Logger._log

    def params_tostring(msg, level):
        level_names = {v: k for k, v in log_level.items()}
        return {"err_msg": msg, "err_type": level_names.get(level, "UNKNOWN"), "err_trace": ""}

    def _fast_py38_handle(
        this: logging.Logger, level, msg, args, exc_info=None, extra=None, stack_info=False, stacklevel=1
    ):
        # 当gunicorn模式启动时，gunicorn会自行上报log信息，也会被此插件拦截，故需剔除gunicorn的log上报 eg：<Logger gunicorn.error (INFO)>
        # uwsgi 不需要拦截，因其无错误log调用
        err_log = this.__str__()
        if "gunicorn" in err_log or "kafka" in err_log or "tornado" in err_log or "werkzeug" in err_log:
            res = _handle(this, level, msg, args, exc_info=None, extra=None, stack_info=False, stacklevel=1)

            return res

        if config.debug and log_level.get(config.logging_level) <= level:
            context = get_context()
            with context.new_local_span(op="Log") as span:
                span.layer = Layer.Local
                span.component = ComponentType.Log

                data = params_tostring(msg, level)
                span.logs.append(Log(items=[LogItem(key=ComponentType.Log, val=data)]))

                res = _handle(this, level, msg, args, exc_info=None, extra=None, stack_info=False, stacklevel=1)

                return res
        else:
            # 当配置文件中的配置的日记级别高于当前上报日志级别，则不做探针上报，也就是说拦截器什么都不做，只负责调用日志原始处理器
            res = _handle(this, level, msg, args, exc_info=None, extra=None, stack_info=False, stacklevel=1)

            return res

    def _fast_py35_handle(this: logging.Logger, level, msg, args, exc_info=None, extra=None):
        # 当gunicorn模式启动时，gunicorn会自行上报log信息，也会被此插件拦截，故需剔除gunicorn的log上报 eg：<Logger gunicorn.error (INFO)>
        # uwsgi 不需要拦截，因其无错误log调用
        err_log = this.__str__()
        if "gunicorn" in err_log or "kafka" in err_log or "tornado" in err_log or "werkzeug" in err_log:
            res = _handle(this, level, msg, args, exc_info=None, extra=None, stack_info=False)

            return res

        if config.debug and log_level.get(config.logging_level) <= level:
            context = get_context()
            with context.new_local_span(op="Log") as span:
                span.layer = Layer.Local
                span.component = ComponentType.Log

                data = params_tostring(msg, level)
                span.logs.append(Log(items=[LogItem(key=ComponentType.Log, val=data)]))

                res = _handle(this, level, msg, args, exc_info=None, extra=None, stack_info=False)

                return res
        else:
            # 当配置文件中的配置的日记级别高于当前上报日志级别，则不做探针上报，也就是说拦截器什么都不做，只负责调用日志原始处理器
            res = _handle(this, level, msg, args, exc_info=None, extra=None, stack_info=False)

            return res

    from sys import version_info

    python_version = version_info[:2]
    if python_version >= (3, 8):
        functions.log("minor python version is:%r", version_info[:2])
        logging.Logger._log = _fast_py38_handle
    else:
        functions.log("python version is:%r", version_info[:2])
        logging.Logger._log = _fast_py35_handle



#!/usr/local/bin python3
# -*- coding: utf-8 -*-

"""
    created by FAST-DEV 2021/4/6
"""
from queue import Queue, Empty

from fast_tracker.agent import Protocol
from fast_tracker.client.udp import UdpTraceSegmentReportService
from fast_tracker.trace.segment import Segment


class UdpProtocol(Protocol):
    def __init__(self):
        self.properties_sent = False
        self.traces_reporter = UdpTraceSegmentReportService()

    def connected(self):
        return True

    def report(self, queue: Queue, block: bool = True):
        def generator():
            while True:
                try:
                    segment = queue.get(block=block)  # type: Segment
                except Empty:
                    return

                # logger.debug('reporting segment %s', segment)

                yield segment

                queue.task_done()

        self.traces_reporter.report(generator=generator())



#!/usr/local/bin python3
# -*- coding: utf-8 -*-

"""
    created by FAST-DEV 2021/4/6
"""

from fast_tracker import Layer, ComponentType
from fast_tracker.trace import tags
from fast_tracker.trace.context import get_context
from fast_tracker.trace.tags import Tag
from fast_tracker.utils import functions


def install():
    from redis.connection import Connection

    _send_command = Connection.send_command

    def _fast_send_command(this: Connection, *args, **kwargs):
        peer = "%s:%s" % (this.host, this.port)
        op = args[0]
        context = get_context()
        with context.new_exit_span(op="Redis/" + op or "/", peer=peer) as span:
            span.layer = Layer.Cache
            span.component = ComponentType.Redis
            res = _send_command(this, *args, **kwargs)
            span.tag(Tag(key=tags.CacheType, val=ComponentType.Redis))
            span.tag(Tag(key=tags.CacheInstance, val=this.db))
            span.tag(Tag(key=tags.CacheCommand, val=op))
            functions.log("redis span is: %r", span)

            return res

    Connection.send_command = _fast_send_command



#!/usr/local/bin python3
# -*- coding: utf-8 -*-

"""
    created by FAST-DEV 2021/4/6
"""


class ServiceManagementClient(object):
    def send_instance_props(self):
        raise NotImplementedError()

    def send_heart_beat(self):
        raise NotImplementedError()


class TraceSegmentReportService(object):
    def report(self, generator):
        raise NotImplementedError()



#!/usr/local/bin python3
# -*- coding: utf-8 -*-

import sys

import os
from setuptools import find_packages

python_version = sys.version_info[:2]

assert python_version in ((2, 7),) or python_version >= (3, 5), "天眼Python探针只支持 Python 2.7 and 3.5+."

with_setuptools = False

try:
    from setuptools import setup

    with_setuptools = True
except ImportError:
    from distutils.core import setup

script_directory = os.path.dirname(__file__)
if not script_directory:
    script_directory = os.getcwd()

readme_file = os.path.join(script_directory, "README.rst")

classifiers = [
    "Development Status :: 4 - Beta",
    "License :: OSI Approved :: Apache Software License",
    "Programming Language :: Python :: 2.7",
    "Programming Language :: Python :: 3.5",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: Implementation :: CPython",
    "Programming Language :: Python :: Implementation :: PyPy",
    "Topic :: System :: Monitoring",
]

kwargs = dict(
    name="fast_tracker",
    #name="fast-trace",
    version="0.2.73",
    description="FAST Python Agent",
    long_description=open(readme_file).read(),
    url="http://doc.mypaas.com.cn/fast/03_%E6%9C%8D%E5%8A%A1%E7%AB%AF%E6%8E%A2%E9%92%88%E9%9B%86%E6%88%90/%E7%AE%80%E4%BB%8B.html",
    author="FAST",
    author_email="yangzy02@mingyuanyun.com",
    maintainer="FAST",
    maintainer_email="yangzy02@mingyuanyun.com",
    license="Apache-2.0",
    # zip_safe=False,
    # classifiers=classifiers,
    # include_package_data=True,
    packages=find_packages(exclude=("tests",)),
    install_requires=[
        "grpcio",
        "grpcio-tools",
        "packaging",
        "gevent",
    ],
    python_requires=">=2.7,!=3.0.*,!=3.1.*,!=3.2.*,!=3.3.*,!=3.4.*",
    package_data={
        "fast_tracker": ["FastTracker.json"],
    },
    license_files = ('apache-skywalking-LICENSE',),
    scripts=["scripts/fast-boot"],
    extras_require={
        "test": [
            "testcontainers",
            "pyyaml",
            "pytest",
        ],
        "http": [
            "requests",
        ],
        "kafka": [
            "kafka",
        ],
    },
)

if with_setuptools:
    kwargs["entry_points"] = {
        "console_scripts": ["fast-admin = fast_tracker.admin:main"],
    }


def with_librt():
    try:
        if sys.platform.startswith("linux"):
            import ctypes.util

            return ctypes.util.find_library("rt")
    except Exception:
        pass


def run_setup():
    kwargs_tmp = dict(kwargs)

    setup(**kwargs_tmp)


run_setup()

print(75 * "*")

print("INFO: Pure Python agent was installed.")


#!/usr/local/bin python3
# -*- coding: utf-8 -*-

"""
    created by iprobeyang@gmail.com 2021/4/12
"""


class ReportStruct:
    def __init__(self):
        # self.product_code = ""  # pc
        # self.app_code = ""  # ac
        self.service_name = ""  # sn
        # self.env_code = ""  # e
        self.tenant_code = ""  # tn
        self.user_code = ""  # u
        self.trace_id = ""  # id
        self.parent_span_id = ""  # p
        self.span_id = ""  # s
        self.peer = ""  # r
        self.span_type = 0  # t
        self.span_layer = 0  # y
        self.start_time = 0  # ts
        self.end_time = 0  # te
        self.duration = 0  # d
        self.component = ""  # c
        self.operation_name = ""  # o
        self.is_error = False  # er
        self.tags = None  # g
        self.logs = None  # l
        self.service_instance = ""  # si
        self.service_version = ""  # sv
        self.tracker_lang = "python"

    @staticmethod
    def convert(obj):
        """
        将数据转成规定格式
        :param obj:
        :return:
        """
        data = {
            # "pc": obj.product_code,
            # "ac": obj.app_code,
            "sn": obj.service_name,
            # "e": obj.env_code,
            "tn": obj.tenant_code,
            "u": obj.user_code,
            "id": obj.trace_id,
            "p": obj.parent_span_id,
            "s": obj.span_id,
            "r": obj.peer,
            "t": obj.span_type,
            "y": obj.span_layer,
            "ts": obj.start_time,
            "te": obj.end_time,
            "d": obj.duration,
            "c": obj.component,
            "o": obj.operation_name,
            "er": obj.is_error,
            "g": obj.tags,
            "l": obj.logs,
            "tv": obj.tracker_version,
            "si": obj.service_instance,
            "sv": obj.service_version,
            "tl": obj.tracker_lang
        }
        if not obj.logs:
            del data["l"]
        if not obj.tags:
            del data["g"]

        return data



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import threading

import os

import time


class Counter:
    def __init__(self):
        self.value = -1

    def next(self) -> int:
        self.value += 1
        return self.value

    def __str__(self):
        return str(self.value)


class SpanCounter:
    def __init__(self):
        self.count = Counter()
        self.value = str(os.getpid()) + "." + str(int(time.time() * 1000)) + "." + str(self.count)

    def next(self) -> str:
        count = self.count.next()
        self.value = str(os.getpid()) + "." + str(int(time.time() * 1000)) + "." + str(count)
        return self.value

    def __str__(self):
        return self.value


class AtomicCounter(Counter):
    def __init__(self):
        super().__init__()
        self._lock = threading.Lock()

    def next(self) -> int:
        with self._lock:
            return Counter.next(self)



#!/usr/local/bin python3
# -*- coding: utf-8 -*-

"""
    created by FAST-DEV 2021/4/6
"""

from fast_tracker import Layer, ComponentType
from fast_tracker.trace import tags
from fast_tracker.trace.context import get_context
from fast_tracker.trace.tags import Tag

version_rule = {"name": "pymongo", "rules": [">=3.7.0"]}


def install():
    from pymongo.bulk import _Bulk
    from pymongo.cursor import Cursor
    from pymongo.pool import SocketInfo

    bulk_op_map = {0: "insert", 1: "update", 2: "delete"}
    # handle insert_many and bulk write
    inject_bulk_write(_Bulk, bulk_op_map)

    # handle find() & find_one()
    inject_cursor(Cursor)

    # handle other commands
    inject_socket_info(SocketInfo)


def inject_socket_info(SocketInfo):
    _command = SocketInfo.command

    def _fast_command(this: SocketInfo, dbname, spec, *args, **kwargs):
        # pymongo sends `ismaster` command continuously. ignore it.
        if spec.get("ismaster") is None:
            address = this.sock.getpeername()
            peer = "%s:%s" % address
            context = get_context()

            operation = list(spec.keys())[0]
            fast_op = operation.capitalize() + "Operation"
            with context.new_exit_span(op="MongoDB/" + fast_op, peer=peer) as span:
                result = _command(this, dbname, spec, *args, **kwargs)

                span.layer = Layer.Database
                span.component = ComponentType.MongoDB
                span.tag(Tag(key=tags.DbType, val="MongoDB"))
                span.tag(Tag(key=tags.DbInstance, val=dbname))

                # get filters
                filters = _get_filter(operation, spec)
                span.tag(Tag(key=tags.DbStatement, val=filters))

        else:
            result = _command(this, dbname, spec, *args, **kwargs)

        return result

    SocketInfo.command = _fast_command


def _get_filter(request_type, spec):
    """
    :param request_type: the request param send to MongoDB
    :param spec: maybe a bson.SON class or a dict
    :return: filter string
    """
    from bson import SON

    if isinstance(spec, SON):
        spec = spec.to_dict()
        spec.pop(request_type)
    elif isinstance(spec, dict):
        spec = dict(spec)
        spec.pop(request_type)

    return request_type + " " + str(spec)


def inject_bulk_write(_Bulk, bulk_op_map):
    _execute = _Bulk.execute

    def _fast_execute(this: _Bulk, *args, **kwargs):
        address = this.collection.database.client.address
        peer = "%s:%s" % address
        context = get_context()

        fast_op = "MixedBulkWriteOperation"
        with context.new_exit_span(op="MongoDB/" + fast_op, peer=peer) as span:
            span.layer = Layer.Database
            span.component = ComponentType.MongoDB

            bulk_result = _execute(this, *args, **kwargs)

            span.tag(Tag(key=tags.DbType, val="MongoDB"))
            span.tag(Tag(key=tags.DbInstance, val=this.collection.database.name))
            filters = ""
            bulk_ops = this.ops
            for bulk_op in bulk_ops:
                opname = bulk_op_map.get(bulk_op[0])
                _filter = opname + " " + str(bulk_op[1])
                filters = filters + _filter + " "

            span.tag(Tag(key=tags.DbStatement, val=filters))

            return bulk_result

    _Bulk.execute = _fast_execute


def inject_cursor(Cursor):
    __send_message = Cursor._Cursor__send_message

    def _fast_send_message(this: Cursor, operation):
        address = this.collection.database.client.address
        peer = "%s:%s" % address

        context = get_context()
        op = "FindOperation"

        with context.new_exit_span(op="MongoDB/" + op, peer=peer) as span:
            span.layer = Layer.Database
            span.component = ComponentType.MongoDB

            # __send_message return nothing
            __send_message(this, operation)

            span.tag(Tag(key=tags.DbType, val=ComponentType.MongoDB))
            span.tag(Tag(key=tags.DbInstance, val=this.collection.database.name))

            filters = "find " + str(operation.spec)
            span.tag(Tag(key=tags.DbStatement, val=filters))

            return

    Cursor._Cursor__send_message = _fast_send_message
